/*

This file is part of VROOM.

Copyright (c) 2015-2021, Julien Coupey.
All rights reserved (see LICENSE).

*/

#include "structures/vroom/solution/computing_times.h"

namespace vroom {

ComputingTimes::ComputingTimes() : loading(0), solving(0), routing(0) {
}

} // namespace vroom
